import random
import string
from typing import Any, Optional, Annotated
import datetime as dt
from fastapi import status
from fastapi.responses import RedirectResponse
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from fastapi import Depends, HTTPException, Request, Response
from sqlalchemy.orm import Session
from sqlalchemy import desc
from passlib.context import CryptContext
from datetime import datetime, timedelta

from api.core.base.services import Service
from api.core.dependencies.email_sender import send_email
from api.db.database import get_db
from api.utils.settings import settings
from api.utils.db_validators import check_model_existence
#from api.v1.models.associations import user_organization_association
from api.v1.models.user import User, roles

from api.v1.models.token_login import TokenLogin
from api.v1.schemas import user
from api.v1.models import Wallet
from api.v1.schemas.wallet import WalletCreate
from api.v1.schemas import token
#from oldies.servicesold.notification_settings import notification_setting_service

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/api/v1/auth/login")
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


class UserService(Service):
    """User service"""

    def fetch_all(
        self, db: Session, page: int, per_page: int, **query_params: Optional[Any]
    ):
        """
        Fetch all users
        Args:
            db: database Session object
            page: page number
            per_page: max number of users in a page
            query_params: params to filter by
        """
        per_page = min(per_page, 10)

        # Enable filter by query parameter
        filters = []
        if all(query_params):
            # Validate boolean query parameters
            for param, value in query_params.items():
                if value is not None and not isinstance(value, bool):
                    raise HTTPException(
                        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                        detail=f"Invalid value for '{param}'. Must be a boolean.",
                    )
                if value == None:
                    continue
                if hasattr(User, param):
                    filters.append(getattr(User, param) == value)
        query = db.query(User)
        total_users = query.count()
        if filters:
            query = query.filter(*filters)
            total_users = query.count()

        all_users: list = (
            query.order_by(desc(User.created_at))
            .limit(per_page)
            .offset((page - 1) * per_page)
            .all()
        )

        return self.all_users_response(all_users, total_users, page, per_page)

    def all_users_response(
        self, users: list, total_users: int, page: int, per_page: int
    ):
        """
        Generates a response for all users
        Args:
            users: a list containing user objects
            total_users: total number of users
        """
        if not users or len(users) == 0:
            return user.AllUsersResponse(
                message="No User(s) for this query",
                status="success",
                status_code=200,
                page=page,
                per_page=per_page,
                total=0,
                data=[],
            )
        all_users = [
            user.UserData.model_validate(usr, from_attributes=True) for usr in users
        ]
        return user.AllUsersResponse(
            message="Users successfully retrieved",
            status="success",
            status_code=200,
            page=page,
            per_page=per_page,
            total=total_users,
            data=all_users,
        )

    def fetch(self, db: Session, id):
        """Fetches a user by their id"""

        user = check_model_existence(db, User, id)

        # return user if user is not deleted
        if not user.is_deleted:
            return user

    def get_user_by_id(self, db: Session, id: str) -> User:
        """Fetches a user by their id"""

        user = check_model_existence(db, User, id)
        return user

    def fetch_by_email(self, db: Session, email):
        """Fetches a user by their email"""

        user = db.query(User).filter(User.email == email).first()

        if not user:
            raise HTTPException(status_code=404, detail="User not found")

        return user

    def create(self, db: Session, schema: user.UserCreate):
        """Creates a new user"""

        

        if db.query(User).filter(User.email == schema.email).first():
            raise HTTPException(
                status_code=400,
                detail="User with this email or username already exists",
            )

        # Hash password
        schema.password = self.hash_password(password=schema.password)

        # Create user object with hashed password and other attributes from schema
        user = User(**schema.model_dump())
        db.add(user)
        db.commit()
        db.refresh(user)

        # # Create notification settings directly for the user
        #notification_setting_service.create(db=db, user=user)

        # create data privacy setting

        #data_privacy = DataPrivacySetting(user_id=user.id)

        #db.add(data_privacy)
        #db.commit()
        #db.refresh(data_privacy)

        return user

    def super_admin_create_user(
        self,
        db: Annotated[Session, Depends(get_db)],
        user_request: user.AdminCreateUser,
    ):
        """
        Creates a user for super admin
        Args:
            db: database Session object
            user_request: The user details to use for creation
        Returns:
            object: the complete details of the newly created user
        """
        try:
            user_exists = (
                db.query(User).filter_by(email=user_request.email).one_or_none()
            )
            if user_exists:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=f"User with {user_request.email} already exists",
                )
            if user_request.password:
                user_request.password = self.hash_password(user_request.password)
            new_user = User(**user_request.model_dump())
            db.add(new_user)
            db.commit()
            db.refresh(new_user)
            user_schema = user.UserData.model_validate(new_user, from_attributes=True)
            return user.AdminCreateUserResponse(
                message="User created successfully",
                status_code=201,
                status="success",
                data=user_schema,
            )
        except Exception as exc:
            db.rollback()
            raise Exception(exc) from exc

    def create_admin(self, db: Session, schema: user.UserCreate):
        """Creates a new admin"""

        del schema.admin_secret

        if db.query(User).filter(User.email == schema.email).first():
            raise HTTPException(
                status_code=400,
                detail="User with this email already exists",
            )

        # Hash password
        schema.password = self.hash_password(password=schema.password)

        # Create user object with hashed password and other attributes from schema
        user = User(**schema.model_dump())
        db.add(user)
        db.commit()
        db.refresh(user)

        # Set user to super admin
        user.is_super_admin = True
        db.commit()

        return user

    def update(self, db: Session, current_user: User, schema: user.UserUpdate, id=None):
        """Function to update a User"""
        # Get user from access token if provided, otherwise fetch user by id
        if db.query(User).filter(User.email == schema.email).first():
            raise HTTPException(
                status_code=400,
                detail="User with this email or username already exists",
            )
        if current_user.is_super_admin and id is not None:
            user = self.fetch(db=db, id=id)
        else:
            user = self.fetch(db=db, id=current_user.id)
        update_data = schema.dict(exclude_unset=True)
        for key, value in update_data.items():
            setattr(user, key, value)
        db.commit()
        db.refresh(user)
        return user
    

    def updatepassword(self, db: Session, current_user: User, schema, id=None):
        """Function to update a User"""
        """# Get user from access token if provided, otherwise fetch user by id
        if db.query(User).filter(User.id == schema.id).first():
            raise HTTPException(
                status_code=400,
                detail="User with this email or username already exists",
            )"""
        if current_user.is_super_admin and id is not None:
            user = self.fetch(db=db, id=id)
        else:
            user = self.fetch(db=db, id=current_user.id)
        update_data = schema.dict(exclude_unset=True)
        for key, value in update_data.items():
            setattr(user, key, value)
        db.commit()
        db.refresh(user)
        return user
    

    def delete(self, db: Session, id=None, access_token: str = Depends(oauth2_scheme)):
        """Function to soft delete a user"""

        # Get user from access token if provided, otherwise fetch user by id
        user = (
            self.get_current_user(access_token, db)
            if id is None
            else check_model_existence(db, User, id)
        )

        user.is_deleted = True
        db.commit()

        return super().delete()

    def authenticate_user(self, db: Session, email: str, password: str):
        """Function to authenticate a user"""

        user = db.query(User).filter(User.email == email).first()

        if not user:
            raise HTTPException(status_code=400, detail="Invalid user credentials")

        if not self.verify_password(password, user.password):
            raise HTTPException(status_code=400, detail="Invalid user credentials")

        return user

    def perform_user_check(self, user: User):
        """This checks if a user is active and verified and not a deleted user"""

        if not user.is_active:
            raise HTTPException(detail="User is not active", status_code=403)

    def hash_password(self, password: str) -> str:
        """Function to hash a password"""

        hashed_password = pwd_context.hash(secret=password)
        return hashed_password

    def verify_password(self, password: str, hash: str) -> bool:
        """Function to verify a hashed password"""

        return pwd_context.verify(secret=password, hash=hash)

    def create_access_token(self, user_id: str) -> str:
        """Function to create access token"""

        expires = dt.datetime.now(dt.timezone.utc) + dt.timedelta(
            minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
        )
        data = {"user_id": user_id, "exp": expires, "type": "access"}
        encoded_jwt = jwt.encode(data, settings.SECRET_KEY, settings.ALGORITHM)
        return encoded_jwt

    def create_refresh_token(self, user_id: str) -> str:
        """Function to create access token"""

        expires = dt.datetime.now(dt.timezone.utc) + dt.timedelta(
            days=settings.JWT_REFRESH_EXPIRY
        )
        data = {"user_id": user_id, "exp": expires, "type": "refresh"}
        encoded_jwt = jwt.encode(data, settings.SECRET_KEY, settings.ALGORITHM)
        return encoded_jwt

    def verify_access_token(self, access_token: str, credentials_exception):
        """Funtcion to decode and verify access token"""

        try:
            payload = jwt.decode(
                access_token, settings.SECRET_KEY, algorithms=[settings.ALGORITHM]
            )
            user_id = payload.get("user_id")
            token_type = payload.get("type")

            if user_id is None:
                raise credentials_exception

            if token_type == "refresh":
                raise HTTPException(detail="Refresh token not allowed", status_code=400)

            token_data = user.TokenData(id=user_id)

        except JWTError as err:
            print(err)
            raise credentials_exception

        return token_data

    def verify_refresh_token(self, refresh_token: str):
        """Funtcion to decode and verify refresh token"""
        credentials_exception = HTTPException(
            status_code=401, detail="Refresh token expired"
        )
        try:
            payload = jwt.decode(
                refresh_token, settings.SECRET_KEY, algorithms=[settings.ALGORITHM]
            )
            user_id = payload.get("user_id")
            token_type = payload.get("type")

            if user_id is None:
                raise credentials_exception

            if token_type == "access":
                raise HTTPException(detail="Access token not allowed", status_code=400)

            token_data = user.TokenData(id=user_id)

        except JWTError:
            raise credentials_exception

        return token_data

    def refresh_access_token(self, current_refresh_token: str):
        """Function to generate new access token and rotate refresh token"""

        credentials_exception = HTTPException(
            status_code=401, detail="Refresh token expired"
        )

        token = self.verify_refresh_token(current_refresh_token)

        if token:
            access = self.create_access_token(user_id=token.id)
            refresh = self.create_refresh_token(user_id=token.id)

            return access, refresh

    def get_current_user(
        self, access_token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)
    ) -> User:
        """Function to get current logged in user"""

        credentials_exception = HTTPException(
            status_code=401,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )

        token = self.verify_access_token(access_token, credentials_exception)
        user = db.query(User).filter(User.id == token.id).first()

        return user

    def deactivate_user(
        self,
        request: Request,
        db: Session,
        schema: user.DeactivateUserSchema,
        user: User,
    ):
        """Function to deactivate a user"""

        if not schema.confirmation:
            raise HTTPException(
                detail="Confirmation required to deactivate account", status_code=400
            )

        self.perform_user_check(user)

        user.is_active = False

        # Create reactivation token
        token = self.create_access_token(user_id=user.id)
        reactivation_link = f"https://{request.url.hostname}/api/v1/users/accounts/reactivate?token={token}"

        # mail_service.send_mail(
        #     to=user.email,
        #     subject='Account deactivation',
        #     body=f'Hello, {user.first_name},\n\nYour account has been deactivated successfully.\nTo reactivate your account if this was a mistake, please click the link below:\n{request.url.hostname}/api/users/accounts/reactivate?token={token}\n\nThis link expires after 15 minutes.'
        # )

        db.commit()

        return reactivation_link

    def reactivate_user(self, db: Session, token: str):
        """This function reactivates a user account"""

        # Validate the token
        try:
            payload = jwt.decode(
                token, settings.SECRET_KEY, algorithms=[settings.ALGORITHM]
            )
            user_id = payload.get("user_id")

            if user_id is None:
                raise HTTPException(400, "Invalid token")

        except JWTError:
            raise HTTPException(400, "Invalid token")

        user = db.query(User).filter(User.id == user_id).first()

        if user.is_active:
            raise HTTPException(400, "User is already active")

        user.is_active = True

        db.commit()

    def change_password(
        self,
        old_password: str,
        new_password: str,
        user: User,
        db: Session,
    ):
        """Endpoint to change the user's password"""

        if not self.verify_password(old_password, user.password):
            raise HTTPException(status_code=400, detail="Incorrect old password")

        user.password = self.hash_password(new_password)
        db.commit()

    def get_current_super_admin(
        self, db: Session = Depends(get_db), token: str = Depends(oauth2_scheme)
    ):
        """Get the current super admin"""
        user = self.get_current_user(db=db, access_token=token)
        if not user.is_super_admin:
            raise HTTPException(
                status_code=403,
                detail="You do not have permission to access this resource",
            )
        return user
    
    def get_current_partner(
        self, db: Session = Depends(get_db), token: str = Depends(oauth2_scheme)
    ):
        """Get the current super admin"""
        user = self.get_current_user(db=db, access_token=token)
        if not user.is_partner:
            raise HTTPException(
                status_code=403,
                detail="You do not have permission to access this resource",
            )
        return user

    def save_login_token(
        self, db: Session, user: User, token: str, expiration: datetime
    ):
        """Save the token and expiration in the user's record"""

        db.query(TokenLogin).filter(TokenLogin.user_id == user.id).delete()

        token = TokenLogin(user_id=user.id, token=token, expiry_time=expiration)
        db.add(token)
        db.commit()
        db.refresh(token)

    def verify_login_token(self, db: Session, schema: token.TokenRequest):
        """Verify the token and email combination"""

        user = db.query(User).filter(User.email == schema.email).first()

        if not user:
            raise HTTPException(status_code=401, detail="Invalid email or token")

        token = db.query(TokenLogin).filter(TokenLogin.user_id == user.id).first()

        if token.token != schema.token or token.expiry_time < datetime.utcnow():
            raise HTTPException(status_code=401, detail="Invalid email or token")

        return user

    def generate_token(self):
        """Generate a 6-digit token"""
        return "".join(
            random.choices(string.digits, k=6)
        ), datetime.utcnow() + timedelta(minutes=10)


    '''def get_users_by_role(self, db: Session, role_id: str, current_user: User):
        """Function to get all users by role"""
        if role_id == "" or role_id is None:
            raise HTTPException(
                status_code=400, 
                detail="Role ID is required"
            )

        user_roles = db.query(user_organization_association).filter(user_organization_association.c.user_id == current_user.id, user_organization_association.c.role.in_(['admin', 'owner'])).all()

        if len(user_roles) == 0:
            raise HTTPException(
                status_code=403, 
                detail="Permission denied. Admin access required."
            )

        users = db.query(User).join(user_organization_association).filter(user_organization_association.c.role == role_id).all()

        if len(users) == 0:
            raise HTTPException(
                status_code=404, 
                detail="No users found for this role"
            )

        return users'''
    
    def create_wallet(self, db: Session, user_id: str) -> Wallet:
        """
        Create a wallet for a user
        Args:
            db: database Session object
            user_id: ID of the user to associate the wallet with
        Returns:
            Wallet: The newly created wallet
        """
        # Ensure the user exists
        user = check_model_existence(db, User, user_id)

        # Check if the user already has a wallet
        existing_wallet = db.query(Wallet).filter(Wallet.user_id == user.id).first()
        if existing_wallet:
            raise HTTPException(
                status_code=400,
                detail="User already has a wallet",
            )

        # Create wallet schema object
        wallet_data = WalletCreate(user_id=user.id)

        # Create the wallet using the schema
        wallet = Wallet(**wallet_data.dict())

        db.add(wallet)
        db.commit()
        db.refresh(wallet)

        return wallet
    
    def fetch_user_refresh_token(self, request, db: Session):
        refresh_token=request.cookies.get("refresh_token")
        if not refresh_token:
            return None
        token_data= user_service.verify_refresh_token(refresh_token)
        user_id =token_data.id
        return user_id
    
    def fetch_user_walletbalance(self, user_id, db: Session):
        
        
        balance = db.query(Wallet).filter(Wallet.user_id == user_id).first()
        
       
        #print(balance.balance, "Owner: ",balance)
        return balance.balance
       
    def change_role(self, current_user, role, db:Session):
        find_user = db.query(User).filter(User.id==current_user.id).first()
        
        find_user.role = role
        db.commit()
        return {
            "message": "Role Switched Successfully",
            "status_code": 200 
        }

    def users_count(self, db:Session):
        count = db.query(User).count()
        return count
    
    
    def create_username(self, db: Session, username, user_id):
        """Creates a new user"""

        
        check_username = db.query(User).filter(User.username == username).first()
        if check_username is None:
          
            user = db.query(User).filter(User.id == user_id).first()
            user.username = username
            db.commit()
        
        if check_username:
            raise HTTPException(
                status_code=400,
                detail="User with this Username already exists",
            )
        
        return {"message": "Username created Successfully"}
        
    
    def about_user(self, db: Session, about, user_id):
        """Update user About"""

        user = db.query(User).filter(User.id == user_id).first()
        user.about_user = about
        

        db.flush()
        db.commit()
        db.refresh(user)

        return {
            "message": "About User Successfully Updated",
            "status_code": 200 
        }


    def fetch_by_username(self, db: Session, username):
        """Fetches a user by their email"""

        user = db.query(User).filter(User.username == username).first()

        if not user:
            raise HTTPException(status_code=404, detail="User not found")

        return user
    

    def get_current_project_admin(
        self, db: Session = Depends(get_db), token: str = Depends(oauth2_scheme)
    ):
        """Get the current super admin"""
        user = self.get_current_user(db=db, access_token=token)
        if user.role != "project":
            raise HTTPException(
                status_code=403,
                detail="You do not have permission to access this resource",
            )
        return user
    
    def system_roles(self, db:Session):
        def convert_enum(enum):
            return {val: val for val in enum.enums}
        
        system_roles = convert_enum(roles)
        return system_roles
     
user_service = UserService()
